# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from copy import copy
from dataclasses import dataclass, field
from typing import Dict, Callable, Any
from .store import InMemoryStore, NotFoundError
from .models import Transaction, Subaddress, PaymentCommand, RefundReason
from .... import jsonrpc, diem_types, txnmetadata, identifier


PENDING_INBOUND_ACCOUNT_ID: str = "pending_inbound_account"


@dataclass
class EventPuller:
    client: jsonrpc.Client
    store: InMemoryStore
    hrp: str
    state: Dict[str, int] = field(default_factory=dict)

    def add(self, address: diem_types.AccountAddress) -> None:
        account = self.client.must_get_account(address)
        self.state[account.received_events_key] = 0

    def fetch(self, process: Callable[[jsonrpc.Event], None], batch_size: int = 100) -> None:
        for key, seq in self.state.items():
            events = self.client.get_events(key, seq, batch_size)
            if events:
                for event in events:
                    process(event)
                    self.state[key] = event.sequence_number + 1

    def head(self) -> None:
        state = None
        while state != self.state:
            state = copy(self.state)
            self.fetch(lambda _: None)

    def save_payment_txn(self, event: jsonrpc.Event) -> None:
        metadata = txnmetadata.decode_structure(event.data.metadata)
        if isinstance(metadata, diem_types.GeneralMetadataV0) and metadata.to_subaddress:
            try:
                res = self.store.find(Subaddress, subaddress_hex=metadata.to_subaddress.hex())
                self._create_txn(res.account_id, event, subaddress_hex=res.subaddress_hex)
            except NotFoundError:
                self._create_txn(PENDING_INBOUND_ACCOUNT_ID, event)
                payee = identifier.encode_account(event.data.sender, metadata.from_subaddress, self.hrp)
                self.store.create(
                    Transaction,
                    account_id=PENDING_INBOUND_ACCOUNT_ID,
                    status=Transaction.Status.pending,
                    currency=event.data.amount.currency,
                    amount=event.data.amount.amount,
                    payee=payee,
                    refund_diem_txn_version=event.transaction_version,
                    refund_reason=RefundReason.invalid_subaddress,
                )
        elif isinstance(metadata, diem_types.TravelRuleMetadataV0) and metadata.off_chain_reference_id:
            cmd = self.store.find(PaymentCommand, reference_id=metadata.off_chain_reference_id)
            self._create_txn(cmd.account_id, event, reference_id=cmd.reference_id)
        elif isinstance(metadata, diem_types.RefundMetadataV0):
            version = int(metadata.transaction_version)
            txn = self.store.find(Transaction, diem_transaction_version=version)
            self._create_txn(
                txn.account_id,
                event,
                refund_diem_txn_version=version,
                refund_reason=RefundReason.from_diem_type(metadata.reason),
            )
        else:
            raise ValueError("unsupported metadata: %s" % metadata)

    def _create_txn(self, account_id: str, event: jsonrpc.Event, **kwargs: Any) -> None:
        self.store.create(
            Transaction,
            account_id=account_id,
            currency=event.data.amount.currency,
            amount=event.data.amount.amount,
            diem_transaction_version=event.transaction_version,
            status=Transaction.Status.completed,
            **kwargs,
        )



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from diem.testing.miniwallet import RestClient, AccountResource, Transaction
from diem import identifier
from typing import Generator
import pytest


amount: int = 123


@pytest.fixture
def sender_account(
    stub_client: RestClient, currency: str, pending_income_account: AccountResource
) -> Generator[AccountResource, None, None]:
    account = stub_client.create_account(balances={currency: amount})
    yield account
    account.log_events()
    # MiniWallet stub saves the payment without account information (subaddress / reference id)
    # into a pending income account before process it.
    # Here we log events of the account for showing more context related to the test
    # when test failed.
    pending_income_account.log_events()


@pytest.fixture
def receiver_account(target_client: RestClient) -> Generator[AccountResource, None, None]:
    account = target_client.create_account()
    yield account
    account.log_events()


def test_receive_payment_with_general_metadata_and_unknown_to_subaddress(
    sender_account: AccountResource, receiver_account: AccountResource, currency: str, hrp: str
) -> None:
    """When received a payment with unknown subaddress, receiver should refund the payment
    by using RefundMetadata with reason `invalid subaddress`.

    Test Plan:
    1. Generate a valid payment URI from receiver account.
    2. Create an invalid payee account identifier by the valid account address and an invalid subaddress.
    3. Send a payment to the invalid payee.
    4. Wait for payment completed.
    5. Assert receiver account does not receive fund.
    6. Assert sender account balances will has same balances before send the payment eventually.
    """

    uri = receiver_account.generate_payment_uri()
    receiver_account_address = uri.intent(hrp).account_address
    invalid_subaddress = identifier.gen_subaddress()
    invalid_payee = identifier.encode_account(receiver_account_address, invalid_subaddress, hrp)
    payment = sender_account.send_payment(currency=currency, amount=amount, payee=invalid_payee)
    wait_for_payment_transaction_complete(sender_account, payment.id)
    assert receiver_account.balance(currency) == 0
    sender_account.wait_for_balance(currency, amount)


def wait_for_payment_transaction_complete(account: AccountResource, payment_id: str) -> None:
    # MiniWallet stub generates `updated_transaction` event when transaction is completed on-chain
    # Payment id is same with Transaction id.
    account.wait_for_event("updated_transaction", status=Transaction.Status.completed, id=payment_id)



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from .app import App, Account, Transaction, PaymentUri, Event, Payment, PaymentCommand, KycSample, falcon_api
from .client import RestClient, AccountResource
from .config import AppConfig, ServerConfig
